#
#  This file is part of bioservices software
#
#  Copyright (c) 2013-2014 - EBI-EMBL
#
#  File author(s):
#      Thomas Cokelaer <cokelaer@ebi.ac.uk>
#
#  Distributed under the GPLv3 License.
#  See accompanying file LICENSE.txt or copy at
#      http://www.gnu.org/licenses/gpl-3.0.html
#
#  website: https://github.com/cokelaer/bioservices
#  documentation: http://packages.python.org/bioservices
#
##############################################################################
"""Interface to the Reactome webs services

.. topic:: What is Reactome?

    :URL: http://www.reactome.org/ReactomeGWT/entrypoint.html
    :Citation: http://www.reactome.org/citation.html
    :REST: http://reactomews.oicr.on.ca:8080/ReactomeRESTfulAPI/RESTfulWS

    .. highlights::

        "REACTOME is an open-source, open access, manually curated and peer-reviewed
        pathway database. Pathway annotations are authored by expert biologists, in
        collaboration with Reactome editorial staff and cross-referenced to many
        bioinformatics databases. These include NCBI Entrez Gene, Ensembl and UniProt
        databases, the UCSC and HapMap Genome Browsers, the KEGG Compound and ChEBI
        small molecule databases, PubMed, and Gene Ontology. ... "

        -- from Reactome web site

"""


from bioservices.services import REST

__all__ = ["Reactome"]


class Reactome:
    """



    .. todo:: interactors, orthology, particiapnts, person,
        query, refernces, schema



    """

    _url = "https://reactome.org/ContentService"

    def __init__(self, verbose=True, cache=False):
        self.services = REST(name="Reactome", url=Reactome._url, verbose="ERROR", cache=False)
        self.debugLevel = verbose

    @property
    def version(self):
        return self.services.http_get("data/database/version", frmt="txt")

    @property
    def name(self):
        return self.services.http_get("data/database/name", frmt="txt")

    def get_discover(self, identifier):
        """The schema.org for an Event in Reactome knowledgebase

        For each event (reaction or pathway) this method generates a
        json file representing the dataset object as defined by
        schema.org (http). This is mainly used by search engines in
        order to index the data

        ::

            r.data_discover("R-HSA-446203")

        """
        res = self.services.http_get("data/discover/{}".format(identifier), frmt="json")
        return res

    def get_diseases(self):
        """list of diseases objects"""
        return self.services.http_get("data/diseases", frmt="json")

    def get_diseases_doid(self):
        """retrieves the list of disease DOIDs annotated in Reactome

        return: dictionary with DOID contained in the values()
        """
        res = self.services.http_get("data/diseases/doid", frmt="txt")
        res = dict([x.split() for x in res.split("\n")])
        return res

    def get_interactors_psicquic_molecule_details(self):
        """Retrieve clustered interaction, sorted by score, of a given accession by resource."""
        raise NotImplementedError

    def get_interactors_psicquic_molecule_summary(self):
        """Retrieve a summary of a given accession by resource"""
        raise NotImplementedError

    def get_interactors_psicquic_resources(self):
        """Retrieve a list of all Psicquic Registries services"""
        raise NotImplementedError

    def get_interactors_static_molecule_details(self):
        """Retrieve a detailed interaction information of a given accession"""
        raise NotImplementedError

    def get_interactors_static_molecule_pathways(self):
        """Retrieve a list of lower level pathways where the interacting molecules can be found"""
        raise NotImplementedError

    def get_interactors_static_molecule_summary(self):
        """Retrieve a summary of a given accession"""
        raise NotImplementedError

    def get_exporter_fireworks(self):
        raise NotImplementedError

    def get_exporter_reaction(self):
        raise NotImplementedError

    def get_exporter_diagram(
        self,
        identifier,
        ext="png",
        quality=5,
        diagramProfile="Modern",
        analysisProfile="Standard",
        filename=None,
    ):
        """Export a given pathway diagram to raster file

        This method accepts identifiers for Event class instances.
        When a diagrammed pathway is provided, the diagram is exported
        to the specified format. When a subpathway is provided, the
        diagram for the parent is exported and the events that are part
        of the subpathways are selected. When a reaction is provided,
        the diagram containing the reaction is exported and the reaction
        is selected.

        :param identifier: Event identifier (it can be a pathway with
            diagram, a subpathway or a reaction)
        :param ext: File extension (defines the image format) in png,
            jpeg, jpg, svg, gif
        :param quality: Result image quality between [1 - 10]. It
            defines the quality of the final image (Default 5)
        :param flg: not implemented
        :param sel: not implemented
        :param diagramProfile: Diagram Color Profile
        :param token: not implemented
        :param analysisProfile: Analysis Color Profile
        :param expColumn: not implemented
        :param filename: if given, save the results in the provided filename

        return: raw data if filename parameter is not set. Otherwise, the data
            is saved in the filename and the function returns None

        """
        assert ext in ["png", "jpg", "jpeg", "svg", "gif"]
        assert quality in range(11)
        assert diagramProfile in ["Modern", "Standard"]
        assert analysisProfile in ["Standard", "Strosobar", "Copper Plus"]

        params = {
            "diagramProfile": diagramProfile,
            "analysisProfile": analysisProfile,
            "quality": quality,
        }

        res = self.services.http_get("exporter/diagram/{}.{}".format(identifier, ext), params=params, frmt=ext)
        if filename:
            if ext != "svg":
                with open(filename, "wb") as fout:
                    fout.write(res)
            else:
                with open(filename, "w") as fout:
                    fout.write(content)
        else:
            return res

    def get_complex_subunits(self, identifier, excludeStructuresSpecifies=False):
        """A list with the entities contained in a given complex

        Retrieves the list of subunits that constitute any given complex.
        In case the complex comprises other complexes, this method
        recursively traverses the content returning each contained
        PhysicalEntity. Contained complexes and entity sets can be
        excluded setting the ‘excludeStructures’ optional parameter to ‘true’

        :param identifier: The complex for which subunits are requested
        :param excludeStructures: Specifies whether contained complexes
            and entity sets are excluded in the response

        ::

            r.get_complex_subunits("R-HSA-5674003")
        """
        params = {"excludeStructuresSpecifies": excludeStructuresSpecifies}
        res = self.services.http_get("data/complex/{}/subunits".format(identifier), params=params, frmt="json")
        return res

    def get_complexes(self, resources, identifier):
        """A list of complexes containing the pair (identifier, resource)

        Retrieves the list of complexes that contain a given (identifier,
        resource). The method deconstructs the complexes into all its
        participants to do so.

        :param resource: The resource of the identifier for complexes are
            requested (e.g. UniProt)
        :param identifier: The identifier for which complexes are requested

        ::

            r.get_complexes(resources, identifier)
            r.get_complexes("UniProt", "P43403")

        """
        res = self.services.http_get("data/complexes/{}/{}".format(resources, identifier), frmt="json")
        return res

    def get_entity_componentOf(self, identifier):
        """A list of larger structures containing the entity

        Retrieves the list of structures (Complexes and Sets) that
        include the given entity as their component. It should be
        mentioned that the list includes only simplified entries
        (type, names, ids) and not full information about each item.

        ::

            r.get_entity_componentOf("R-HSA-199420")

        """
        res = self.services.http_get("data/entity/{}/componentOf".format(identifier), frmt="json")
        return res

    def get_entity_otherForms(self, identifier):
        """All other forms of PhysicalEntity

        Retrieves a list containing all other forms of the given
        PhysicalEntity. These other forms are PhysicalEntities that
        share the same ReferenceEntity identifier, e.g. PTEN
        H93R[R-HSA-2318524] and PTEN C124R[R-HSA-2317439] are two
        forms of PTEN.

        ::

            r.get_entity_otherForms("R-HSA-199420")

        """
        res = self.services.http_get("data/entity/{}/otherForms".format(identifier), frmt="json")
        return res

    def get_event_ancestors(self, identifier):
        """The ancestors of a given event

        The Reactome definition of events includes pathways and reactions.
        Although events are organised in a hierarchical structure, a single
        event can be in more than one location, i.e. a reaction can take
        part in different pathways while, in the same way, a sub-pathway
        can take part in many pathways. Therefore, this method retrieves
        a list of all possible paths from the requested event to the top
        level pathway(s).

        :param identifier: The event for which the ancestors are requested

        ::

            r.get_event_ancestors("R-HSA-5673001")

        """
        res = self.services.http_get("data/event/{}/ancestors".format(identifier), frmt="json")
        return res

    def get_eventsHierarchy(self, species):
        """The full event hierarchy for a given species

        Events (pathways and reactions) in Reactome are organised in a
        hierarchical structure for every species. By following all
        ‘hasEvent’ relationships, this method retrieves the full event
        hierarchy for any given species. The result is a list of tree
        structures, one for each TopLevelPathway. Every event in these trees is
        represented by a PathwayBrowserNode. The latter contains the stable identifier,
        the name, the species, the url, the type, and the diagram of the particular
        event.

        :param species: Allowed species filter: SpeciesName (eg: Homo sapiens)
            SpeciesTaxId (eg: 9606)

        ::

            r.get_eventsHierarchy(9606)
        """

        res = self.services.http_get("data/eventsHierarchy/{}".format(species), frmt="json")
        return res

    def get_exporter_sbml(self, identifier):
        """Export given Pathway to SBML


        :param identifier: DbId or StId of the requested database object

        ::

            r.exporter_sbml("R-HSA-68616")

        """
        res = self.services.http_get("exporter/sbml/{}.xml".format(identifier), frmt="xml")
        return res

    def get_pathway_containedEvents(self, identifier):
        """All the events contained in the given event

        Events are the building blocks used in Reactome to represent
        all biological processes, and they include pathways and reactions.
        Typically, an event can contain other events. For example, a
        pathway can contain smaller pathways and reactions. This method
        recursively retrieves all the events contained in any given event.

        ::

            res = r.get_pathway_containedEvents("R-HSA-5673001")

        """
        res = self.services.http_get("data/pathway/{}/containedEvents".format(identifier), frmt="json")
        return res

    def get_pathway_containedEvents_by_attribute(self, identifier, attribute):
        """A single property for each event contained in the given event

        Events are the building blocks used in Reactome to represent all
        biological processes, and they include pathways and reactions.
        Typically, an event can contain other events. For example, a
        pathway can contain smaller pathways (subpathways) and reactions.
        This method recursively retrieves a single attribute for each of
        the events contained in the given event.


        :param identifier: The event for which the contained events are requested
        :param attribute: Attrubute to be filtered

        ::

             r.get_pathway_containedEvents_by_attribute("R-HSA-5673001", "stId")

        """
        res = self.services.http_get(
            "data/pathway/{}/containedEvents/{}".format(identifier, attribute),
            frmt="txt",
        )
        try:
            res = [x.strip() for x in res[1:-1].split(",")]
        except:
            pass
        return res

    def get_pathways_low_diagram_entity(self, identifier):
        """A list of lower level pathways with diagram containing
        a given entity or event

        This method traverses the event hierarchy and retrieves the
        list of all lower level pathways that have a diagram and
        contain the given PhysicalEntity or Event.

        :param identifier: The entity that has to be present in the pathways
        :param species:  The species for which the pathways are requested.
            Taxonomy identifier (eg: 9606) or species name (eg: ‘Homo sapiens’)

        ::

            r.get_pathways_low_diagram_entity("R-HSA-199420")

        """
        res = self.services.http_get("data/pathways/low/diagram/entity/{}".format(identifier), frmt="json")
        return res

    def get_pathways_low_diagram_entity_allForms(self, identifier):
        """

        ::

            r.get_pathways_low_diagram_entity_allForms("R-HSA-199420")
        """
        res = self.services.http_get(
            "data/pathways/low/diagram/entity/{}/allForms".format(identifier),
            frmt="json",
        )
        return res

    def get_pathways_low_entity(self, identifier):
        """A list of lower level pathways containing a given entity or event

        This method traverses the event hierarchy and retrieves the
        list of all lower level pathways that contain the given
        PhysicalEntity or Event.

        ::

            r.get_pathways_low_entity("R-HSA-199420")
        """
        res = self.services.http_get("data/pathways/low/entity/{}".format(identifier), frmt="json")
        return res

    def get_pathways_low_entity_allForms(self, identifier):
        """A list of lower level pathways containing any form of a given entity

        This method traverses the event hierarchy and retrieves the list of all
        lower level pathways that contain the given PhysicalEntity in any of
        its variant forms. These variant forms include for example different
        post-translationally modified versions of a single protein, or the
        same chemical in different compartments.

        ::

            r.get_pathways_low_entity_allForms("R-HSA-199420")
        """
        res = self.services.http_get("data/pathways/low/entity/{}/allForms".format(identifier), frmt="json")
        return res

    def get_pathways_top(self, species):
        res = self.services.http_get("data/pathways/top/{}".format(species), frmt="json")
        return res

    def get_references(self, identifier):
        """All referenceEntities for a given identifier

        Retrieves a list containing all the reference entities for a given
        identifier.

        ::

            r.get_references(15377)

        """
        res = self.services.http_get("references/mapping/{}".format(identifier), frmt="json")
        return res

    def get_mapping_identifier_pathways(self, resource, identifier):
        res = self.services.http_get("data/mapping/{}/{}/pathways".format(resource, identifier), frmt="json")
        return res

    def get_mapping_identifier_reactions(self, resource, identifier):
        res = self.services.http_get("data/mapping/{}/{}/reactions".format(resource, identifier), frmt="json")

    def search_facet(self):
        """A list of facets corresponding to the whole Reactome search data

        This method retrieves faceting information on the whole Reactome search data.


        """
        res = self.services.http_get("search/facet", frmt="json")
        return res

    def search_facet_query(self, query):
        """A list of facets corresponding to a specific query

        This method retrieves faceting information on a specific query

        """
        res = self.services.http_get("search/facet_query?query={}".format(query), frmt="json")
        return res

    def search_query(self, query):
        """Queries Solr against the Reactome knowledgebase

        This method performs a Solr query on the Reactome knowledgebase.
        Results can be provided in a paginated format.

        """
        res = self.services.http_get("search/query?query={}".format(query), frmt="json")
        return res

    def search_spellcheck(self, query):
        """Spell-check suggestions for a given query

        This method retrieves a list of spell-check suggestions
        for a given search term.

        """
        res = self.services.http_get("search/spellcheck?query={}".format(query), frmt="json")
        return res

    def search_suggest(self, query):
        """Autosuggestions for a given query


        This method retrieves a list of suggestions for a given search term.

        ::

            >>> r.http_get("search/suggest?query=apopt")
            ['apoptosis', 'apoptosome', 'apoptosome-mediated', 'apoptotic']

        """
        res = self.services.http_get("search/suggest?query={}".format(identifier), frmt="json")
        return res

    def get_species_all(self):
        """the list of all species in Reactome"""
        res = self.services.http_get("data/species/all", frmt="json")
        return res

    def get_species_main(self):
        """the list of main species in Reactome

        ::

            r.get_species_main()


        """
        res = self.services.http_get("data/species/main", frmt="json")
        return res


"""
class ReactomePathway(object):

    def __init__(self, entry):
        self.raw_data = copy.deepcopy(entry)
        # just adding the attributes to make life easier
        for k in self.raw_data._keyord:
            setattr(self, k, getattr(self.raw_data, k))

    def __str__(self):

        txt = "id: " + str(self.id)
        txt += "\nname: " + str(self.name)

        txt += "\nhasComponent:"
        if self.hasComponent:
            txt += str(len(self.hasComponent))

        if self.raw_data.compartment:
            txt += "\ncompartment:" + str(len(self.raw_data.compartment))
        else:
            txt += "\ncompartment: " + str(self.raw_data.compartment)

        txt += "\nliteratureReference:"
        this = self.raw_data.literatureReference
        if this:
            for i,x in enumerate(this):
                txt += " --- %s: %s %s\n" % (i, x.id, x.name)

        txt += "\nspecies:"
        this = self.raw_data.species
        if this:
            txt += "  %s (%s)" % (this.scientificName, this.id)

        txt += "\nsummation:"
        this = self.raw_data.summation
        if this:
            for x in this:
                txt += " - %s \n" % (self.raw_data.id)

        txt += "\ngoBiologicalProcess:"
        this = self.raw_data.goBiologicalProcess
        if this: txt += "\n - %s %s\n" % (this.id, this.name)

        txt += "\ninferredFrom:" + str(self.raw_data.inferredFrom)
        txt += "\northologousEvent: "
        this = self.raw_data.orthologousEvent
        if this:
            txt += str(len(this))

        txt += "\nprecedingEvent: " + str(self.raw_data.precedingEvent)
        txt += "\nevidence Type: " + str(self.raw_data.evidenceType) + "\n"

        return txt
"""



#
#  This file is part of bioservices software
#
#  Copyright (c) 2013-2023 - EBI-EMBL
#
#  File author(s):
#      Thomas Cokelaer <cokelaer@ebi.ac.uk>
#
#
#  Distributed under the GPLv3 License.
#  See accompanying file LICENSE.txt or copy at
#      http://www.gnu.org/licenses/gpl-3.0.html
#
#  website: https://github.com/cokelaer/bioservices
#  documentation: http://packages.python.org/bioservices
#
##############################################################################
"""Interface to some part of the UniProt web service

.. topic:: What is COG service?

    :URL: https://www.ncbi.nlm.nih.gov/research/cog/webservices/
    :Citation:

    .. highlights::

        Database of Clusters of Orthologous Genes (COGs)

        -- From COG web site, Jan 2021


"""
import types
import io
import sys

from tqdm import tqdm

import pandas as pd

from bioservices.services import REST
from bioservices import logger

logger.name = __name__


__all__ = ["COG"]


class COG:
    """Interface to the COG service

    Note that in addition to the original COG service from NCBI, this interface also
    helps you in searching for orgamism, and retrieve all pages in a single command
    (rather than scanning yourself all pages).

    Here is an example of getting the COG for ecoli. Your first the exact matching name.
    Bioservices provices a function to serch for the exact organism name that will be understood
    by the COG service (here Escherichia_coli_K-12_sub_MG1655 ... you cannot guess it really)
    ::

        from bioservices import COG
        c = COG()
        c.search_organism('coli')

        # the output of the previous command gives you the name
        c.get_cogs_by_orgnanism('Escherichia_coli_K-12_sub_MG1655')
    """

    _url = "https://www.ncbi.nlm.nih.gov/research/cog/api"

    def __init__(self, verbose=False, cache=False):
        """**Constructor**"""
        self.services = REST(name="cog", url=COG._url, verbose=verbose, cache=cache)
        self.show_progress = True

    def _get_all(self, service_name="cog", params={}):
        page = 1
        params["page"] = page
        res = self.services.http_get(service_name, frmt="json", params=params)
        total = res["count"]

        pbar = tqdm(total=total, disable=not self.show_progress, leave=False)

        # sometimes, a 404 is returned, let us try several times.
        trials = 3

        while True:
            params["page"] += 1
            for _ in range(trials):
                other = self.services.http_get(service_name, frmt="json", params=params)
                try:
                    res["results"].extend(other["results"])
                    break
                except TypeError:
                    pass
                except Exception as err:
                    raise (err)
            pbar.update(len(other["results"]))
            if other["next"] is None:
                break
        pbar.close()

        return res

    def get_cogs(self, **kwargs):
        """Get COGs. Unfortunately, the API sends 10 COGS at a time given a
        specific page.

        The dictionary returned contains the results, count, previous and next
        page.
        """
        if kwargs.get("page") is None:
            res = self._get_all("cog", params=kwargs)
        else:
            res = self.services.http_get("cog", frmt="json", params=kwargs)
        return res

    def get_cogs_by_gene(self, gene, page=None):
        """Filter COGs by gene tag: MK0280"""
        return self.get_cogs(**{"gene": gene, "page": page})

    def get_cogs_by_id(self, cog_id, page=None):
        """Filter COGs by COG ID tag: COG0003"""
        return self.get_cogs(**{"cog": cog_id, "page": page})

    def get_cogs_by_assembly_id(self, assembly_id, page=None):
        """Filter COGs by assembly ID: GCA_000007185.1"""
        return self.get_cogs(**{"assembly": assembly_id, "page": page})

    def get_cogs_by_organism(self, name, page=None):
        """Filter COGs by organism name: Nitrosopumilus_maritimus_SCM1"""
        return self.get_cogs(**{"organism": name, "page": page})

    def get_cogs_by_taxon_id(self, taxon_id, page=None):
        """Filter COGs by taxid: 1229908"""
        return self.get_cogs(**{"taxid": taxon_id, "page": page})

    def get_cogs_by_category(self, category, page=None):
        """Filter COGs by Taxonomic Category: ACTINOBACTERIA"""
        return self.get_cogs(**{"category": category, "page": page})

    def get_cogs_by_category_id(self, category, page=None):
        """Filter COGs by Taxonomic Category taxid: 651137"""
        return self.get_cogs(**{"cat_taxid": category, "page": page})

    def get_cogs_by_protein_name(self, protein, page=None):
        """Filter COGs by Protein name: AJP49128.1"""
        return self.get_cogs(**{"protein": protein, "page": page})

    def get_cogs_by_id_and_category(self, cog_id, category, page=None):
        """Filter COGs by COG id and Taxonomy Categories: COG0004 and CYANOBACTERIA"""
        return self.get_cogs(**{"cog": cog_id, "category": category, "page": page})

    def get_cogs_by_id_and_organism(self, cog_id, organism, page=None):
        """Filter COGs by COG id and organism: COG0004 and Escherichia_coli_K-12_sub_MG1655"""
        return self.get_cogs(**{"cog": cog_id, "organism,": organism, "page": page})

    def get_all_cogs_definition(self, page=None):
        """Get all COG Definitions:"""
        if page is None:
            res = self._get_all("cogdef")
        else:
            res = self.services.http_get("cogdef", frmt="json", params={"page": page})
        return res

    def get_cog_definition_by_cog_id(self, cog_id):
        """Get specific COG Definitions by COG: COG0003"""
        return self.services.http_get("cogdef", frmt="json", params={"cog": cog_id})

    def get_cog_definition_by_name(self, cog, page=None):
        """Get specific COG Definitions by name: Thiamin-binding stress-response protein YqgV, UPF0045 family"""

        if page is None:
            res = self._get_all("cogdef", params={"name": cog})
        else:
            res = self.services.http_get("cogdef", frmt="json", params={"name": cog})
        return res

    def get_taxonomic_categories(self, page=None):
        """Get all Taxonomic Categories.

        if page is set, only that page is returned. There are 10 entires per page.
        if page is unset (default), all results are returned.


        ::

            from bioservices import COG
            c = COG()
            names = [x['name'] for x in c.get_taxonomic_categories()['results']]

        """
        if page is None:
            res = self._get_all("taxonomy", params={})
        else:
            res = self.services.http_get("taxonomy", frmt="json", params={"page": page})

        return res

    def get_taxonomic_category_by_name(self, name, page=None):
        """Get specific Taxonomic Category by name


        c.get_taxonomic_category_by_name("ALPHAPROTEOBACTERIA")
        """
        if page is None:
            res = self._get_all("taxonomy", params={"name": name})
        else:
            res = self.services.http_get("taxonomy", frmt="json", params={"name": name, "page": page})
        return res

    def search_organism(self, name):
        """Return candidates that match the input name.

        :param str name:
        :return: list of items. Each item is a dictionary with genome name, assembly identifier and taxon identifier.

        """
        results = self.get_taxonomic_categories()
        candidates = []
        for x in results["results"]:
            for y in x["organisms"]:
                if name in y["genome_name"].lower():
                    candidates.append(y)
        return candidates



#
#  This file is part of bioservices software
#
#  Copyright (c) 2013-2020 - EBI-EMBL - Institut Pasteur
#
#  File author(s):
#      Thomas Cokelaer <cokelaer@ebi.ac.uk>
#      Thomas Cokelaer <thomas.cokelaer@pasteur.fr>
#
#
#  Distributed under the GPLv3 License.
#  See accompanying file LICENSE.txt or copy at
#      http://www.gnu.org/licenses/gpl-3.0.html
#
#  website: https://github.com/cokelaer/bioservices
#  documentation: http://bioservices.readthedocs.io
#
##############################################################################
"""Interface to the PDBe web Service (v2 API).

.. topic:: What is PDBe ?

    :URL: https://www.ebi.ac.uk/pdbe/
    :REST: https://www.ebi.ac.uk/pdbe/api/doc/

    .. highlights::

        PDBe is a founding member of the Worldwide Protein Data Bank which
        collects, organises and disseminates data on biological macromolecular
        structures. In collaboration with the other Worldwide Protein Data Bank (wwPDB)
        partners, we work to collate, maintain and provide access to the global
        repository of macromolecular structure models, the Protein Data Bank (PDB).

        -- PDBe home page, June 2020

"""

import json as _json

from bioservices.services import REST

__all__ = ["PDBe"]


class PDBe:
    """Interface to part of the `PDBe <http://www.ebi.ac.uk/pdbe>`_ service

    .. doctest::

        >>> from bioservices import PDBe
        >>> s = PDBe()
        >>> res = s.get_files("1FBV")

    """

    _entry_prefix = "pdb/entry"

    def __init__(self, verbose=False, cache=False):
        """.. rubric:: Constructor

        :param bool verbose: prints informative messages (default is off)

        """
        url = "https://www.ebi.ac.uk/pdbe/api/v2"
        self.services = REST(name="PDBe", url=url, verbose=verbose, cache=cache, url_defined_later=True)

    def _check_id(self, pdbid):
        if isinstance(pdbid, list):
            pdbid = ",".join(pdbid)

        if isinstance(pdbid, str):
            for item in pdbid.split(","):
                assert len(item) == 4, "a 4-character PDB id code is required"
        else:
            raise TypeError(
                "pdb id must be either a 4-character pdb id, a list of valid PDB ids, or a string made of pdb ids, separated by commas"
            )

        return pdbid

    def _return(self, res):
        if res in (404, 410):
            return {}
        return res

    def _post_json(self, endpoint, query):
        """Send a POST request with JSON content type to the v2 API.

        The v2 API expects POST bodies as JSON-encoded strings containing
        comma-separated PDB IDs.
        """
        headers = self.services.get_headers("json")
        return self.services.http_post(
            endpoint,
            data=_json.dumps(query),
            frmt="json",
            headers=headers,
        )

    def _get_or_post(self, endpoint, query):
        """Handle single PDB ID via GET or multiple PDB IDs via POST.

        :param endpoint: the API endpoint name (e.g. 'summary', 'molecules')
        :param query: a PDB id, a comma-separated string of PDB ids, or a list of PDB ids
        :returns: the API response
        """
        query = self._check_id(query)
        if "," not in query:
            res = self.services.http_get("{}/{}/{}".format(self._entry_prefix, endpoint, query))
        else:
            res = self._post_json("{}/{}".format(self._entry_prefix, endpoint), query)
        return self._return(res)

    def get_summary(self, query):
        """Returns summary of a PDB entry

        This can be title of the entry, list of depositors, date of deposition,
        date of release, date of latest revision, experimental method, list
        of related entries in case split entries, etc.

        :param query: a 4-character PDB id code

        ::

            p.get_summary('1cbs')
            p.get_summary('1cbs,2kv8')
            p.get_summary(['1cbs', '2kv8'])

        """
        return self._get_or_post("summary", query)

    def get_molecules(self, query):
        """Return details of molecules (or entities in mmcif-speak) modelled in the entry

        This can be entity id, description, type, polymer-type (if applicable), number
        of copies in the entry, sample preparation method, source organism(s)
        (if applicable), etc.

        :param query: a 4-character PDB id code

        ::

            p.get_molecules('1cbs')
            p.get_molecules('1cbs,2kv8')

        """
        return self._get_or_post("molecules", query)

    def get_entities(self, query):
        """Return details of entities modelled in the entry

        This is an alias for :meth:`get_molecules` using the ``entities`` endpoint.

        :param query: a 4-character PDB id code

        ::

            p.get_entities('1cbs')
            p.get_entities('1cbs,2kv8')

        """
        return self._get_or_post("entities", query)

    def get_publications(self, query):
        """Return publications associated with the entry

        Provides details of publications associated with an entry, such as title
        of the article, journal name, year of publication, volume, pages, doi,
        pubmed_id, etc. Primary citation is listed first.

        :param query: a 4-character PDB id code

        ::

            p.get_publications('1cbs')
            p.get_publications('1cbs,2kv8')

        """
        return self._get_or_post("publications", query)

    def get_related_publications(self, query):
        """Return publications obtained from both EuroPMC and UniProt.

        These are articles which cite the primary citation of the entry, or
        open-access articles which mention the entry id without explicitly citing the
        primary citation of an entry.

        :param query: a 4-character PDB id code

        ::

            p.get_related_publications('1cbs')
            p.get_related_publications('1cbs,2kv8')

        """
        return self._get_or_post("related_publications", query)

    def get_experiment(self, query):
        """Provides details of experiment(s) carried out in determining the structure of the entry.

        Each experiment is described in a separate dictionary.
        For X-ray diffraction, the description consists of resolution, spacegroup, cell
        dimensions, R and Rfree, refinement program, etc.
        For NMR, details of spectrometer, sample, spectra, refinement, etc. are
        included.
        For EM, details of specimen, imaging, acquisition, reconstruction, fitting etc.
        are included.

        :param query: a 4-character PDB id code

        ::

            p.get_experiment('1cbs')
            p.get_experiment('1cbs,2kv8')

        """
        return self._get_or_post("experiment", query)

    def get_ligand_monomers(self, query):
        """Provides a list of modelled instances of ligands,

        i.e. 'bound' molecules that are not waters.

        :param query: a 4-character PDB id code

        ::

            p.get_ligand_monomers('1cbs')
            p.get_ligand_monomers('1cbs,2kv8')

        """
        return self._get_or_post("ligand_monomers", query)

    def get_modified_residues(self, query):
        """Provides a list of modelled instances of modified amino acids or
        nucleotides in protein, DNA or RNA chains.

        :param query: a 4-character PDB id code

        ::

            p.get_modified_residues('4v5j')
            p.get_modified_residues('4v5j,1cbs')

        """
        return self._get_or_post("modified_AA_or_NA", query)

    def get_mutated_residues(self, query):
        """Provides a list of modelled instances of mutated amino acids or
        nucleotides in protein, DNA or RNA chains.

        :param query: a 4-character PDB id code

        ::

            p.get_mutated_residues('1bgj')
            p.get_mutated_residues('1bgj,4v5j')

        """
        return self._get_or_post("mutated_AA_or_NA", query)

    def get_release_status(self, query):
        """Provides status of a PDB entry (released, obsoleted, on-hold etc)
        along with some other information such as authors, title, experimental method,
        etc.

        :param query: a 4-character PDB id code

        ::

            p.get_release_status('1cbs')
            p.get_release_status('1cbs,4v5j')

        """
        return self._get_or_post("status", query)

    def get_observed_ranges(self, query):
        """Provides observed ranges, i.e., segments of structural coverage of
        polymeric molecules that are modelled fully or partly.

        :param query: a 4-character PDB id code

        ::

            p.get_observed_ranges('1cbs')
            p.get_observed_ranges('1cbs,4v5j')

        """
        return self._get_or_post("polymer_coverage", query)

    def get_observed_ranges_in_pdb_chain(self, query, chain_id):
        """Provides observed ranges, i.e., segments of structural coverage of
        polymeric molecules in a particular chain.

        :param query: a 4-character PDB id code
        :param chain_id: a PDB chain ID

        ::

            p.get_observed_ranges_in_pdb_chain('1cbs', 'A')

        """
        query = self._check_id(query)
        res = self.services.http_get(
            "{}/polymer_coverage/{}/chain/{}".format(self._entry_prefix, query, chain_id)
        )
        return self._return(res)

    def get_secondary_structure(self, query):
        """Provides residue ranges of regular secondary structure

        (alpha helices and beta strands) found in protein chains of the entry.
        For strands, sheet id can be used to identify a beta sheet.

        :param query: a 4-character PDB id code

        ::

            p.get_secondary_structure('1cbs')
            p.get_secondary_structure('1cbs,4v5j')

        """
        return self._get_or_post("secondary_structure", query)

    def get_residue_listing(self, query):
        """Lists all residues (modelled or otherwise) in the entry.

        Except waters, along with details of the fraction of expected atoms modelled for
        the residue and any alternate conformers.

        :param query: a 4-character PDB id code

        ::

            p.get_residue_listing('1cbs')

        """
        query = self._check_id(query)
        res = self.services.http_get(
            "{}/residue_listing/{}".format(self._entry_prefix, query)
        )
        return self._return(res)

    def get_residue_listing_in_pdb_chain(self, query, chain_id):
        """Lists all residues (modelled or otherwise) in a particular chain.

        Except waters, along with details of the fraction of expected atoms
        modelled for the residue and any alternate conformers.

        :param query: a 4-character PDB id code
        :param chain_id: a PDB chain ID

        ::

            p.get_residue_listing_in_pdb_chain('1cbs', 'A')

        """
        query = self._check_id(query)
        res = self.services.http_get(
            "{}/residue_listing/{}/chain/{}".format(self._entry_prefix, query, chain_id)
        )
        return self._return(res)

    def get_binding_sites(self, query, entity_id):
        """Provides details on binding sites for a specific entity in the entry.

        STRUCT_SITE records in PDB files (or mmcif equivalent thereof), such as ligand,
        residues in the site, description of the site, etc.

        :param query: a 4-character PDB id code
        :param entity_id: an entity ID (integer or string)

        ::

            p.get_binding_sites('1cbs', 1)

        """
        query = self._check_id(query)
        res = self.services.http_get(
            "{}/binding_sites/{}/{}".format(self._entry_prefix, query, entity_id)
        )
        return self._return(res)

    def get_files(self, query):
        """Provides URLs and brief descriptions (labels) for PDB entry

        Also, for mmcif files, biological assembly files, FASTA file for sequences,
        SIFTS cross reference XML files, validation XML files, X-ray structure
        factor file, NMR experimental constraints files, etc.

        :param query: a 4-character PDB id code

        ::

            p.get_files('1cbs')
            p.get_files('1cbs,4v5j')

        """
        return self._get_or_post("files", query)

    def get_observed_residues_ratio(self, query):
        """Provides the ratio of observed residues for each chain in each molecule.

        The list of chains within an entity is sorted by observed_ratio (descending order),
        partial_ratio (ascending order), and number_residues (descending order).

        :param query: a 4-character PDB id code

        ::

            p.get_observed_residues_ratio('1cbs')
            p.get_observed_residues_ratio('1cbs,4v5j')

        """
        return self._get_or_post("observed_residues_ratio", query)

    def get_assembly(self, query):
        """Provides information for each assembly of a given PDB ID.

        This information is broken down at the entity level for each assembly. The
        information given includes the molecule name, type and class, the chains where
        the molecule occur, and the number of copies of each entity in the assembly.

        :param query: a 4-character PDB id code

        ::

            p.get_assembly('1cbs')
            p.get_assembly('1cbs,4v5j')

        """
        return self._get_or_post("assembly", query)

    def get_electron_density_statistics(self, query):
        """Provides statistics for electron density.

        :param query: a 4-character PDB id code

        ::

            p.get_electron_density_statistics('1cbs')
            p.get_electron_density_statistics('1cbs,4v5j')

        """
        return self._get_or_post("electron_density_statistics", query)

    def get_functional_annotation(self, query):
        """Provides functional annotation of all ligands, i.e. 'bound' molecules.

        :param query: a 4-character PDB id code

        ::

            p.get_functional_annotation('1cbs')

        """
        query = self._check_id(query)
        res = self.services.http_get(
            "{}/cofactor/{}".format(self._entry_prefix, query)
        )
        return self._return(res)

    def get_drugbank_annotation(self, query):
        """Provides DrugBank annotation of all ligands, i.e. 'bound' molecules.

        :param query: a 4-character PDB id code

        ::

            p.get_drugbank_annotation('5hht')

        """
        query = self._check_id(query)
        res = self.services.http_get(
            "{}/drugbank/{}".format(self._entry_prefix, query)
        )
        return self._return(res)

    def get_related_dataset(self, query):
        """Provides DOIs for related raw experimental datasets.

        Includes diffraction image data, small-angle scattering data and
        electron micrographs.

        :param query: a 4-character PDB id code

        ::

            p.get_related_dataset('5o8b')
            p.get_related_dataset('5o8b,5o8b')

        """
        return self._get_or_post("related_experiment_data", query)

    def get_branched_entities(self, query):
        """Provides data for branched carbohydrate entities within an entry.

        Overall information about each unique branched carbohydrate is returned,
        along with detailed information about each carbohydrate monomer within
        the branched entity.

        :param query: a 4-character PDB id code

        ::

            p.get_branched_entities('3d12')
            p.get_branched_entities('3d12,7v7u')

        """
        return self._get_or_post("branched", query)

    def get_carbohydrate_polymer(self, query):
        """Provides data for carbohydrate polymers within an entry.

        :param query: a 4-character PDB id code

        ::

            p.get_carbohydrate_polymer('3d12')
            p.get_carbohydrate_polymer('3d12,7v7u')

        """
        return self._get_or_post("carbohydrate_polymer", query)



#
#  This file is part of bioservices software
#
#  Copyright (c) 2013-2014 - EBI-EMBL
#
#  Distributed under the GPLv3 License.
#  See accompanying file LICENSE.txt or copy at
#      http://www.gnu.org/licenses/gpl-3.0.html
#
#  website: https://github.com/cokelaer/bioservices
#  documentation: http://packages.python.org/bioservices
#
##############################################################################
"""This module provides a class :class:`ChEBI`

.. topic:: What is ChEBI

    :URL:  https://www.ebi.ac.uk/chebi/init.do
    :REST: https://www.ebi.ac.uk/chebi/backend/api/public


    .. highlights::

        "The database and ontology of Chemical Entities of Biological Interest

        -- From ChEBI web page June 2013


"""
from bioservices import REST
from bioservices import logger

logger.name = __name__

__all__ = ["ChEBI"]

# Mapping from old SOAP relation type names to new REST API relation type names
_RELATION_TYPE_MAP = {
    "is a": "is_a",
    "has part": "has_part",
    "has role": "has_role",
    "is conjugate base of": "is_conjugate_base_of",
    "is conjugate acid of": "is_conjugate_acid_of",
    "is tautomer of": "is_tautomer_of",
    "is enantiomer of": "is_enantiomer_of",
    "has functional parent": "has_functional_parent",
    "has parent hybride": "has_parent_hydride",
    "is substituent group of": "is_substituent_group_from",
}


class ChebiEntity(dict):
    """A dict subclass returned by ChEBI REST API calls.

    Provides attribute-style access to common compound fields for
    backward compatibility with the old SOAP-based interface.
    """

    @property
    def mass(self):
        """Molecular mass of the compound."""
        chem = self.get("chemical_data") or {}
        return chem.get("mass")

    @property
    def smiles(self):
        """SMILES string of the default structure."""
        struct = self.get("default_structure") or {}
        return struct.get("smiles")

    @property
    def inchiKey(self):
        """Standard InChI key of the default structure."""
        struct = self.get("default_structure") or {}
        return struct.get("standard_inchi_key")

    @property
    def formula(self):
        """Molecular formula."""
        chem = self.get("chemical_data") or {}
        return chem.get("formula")

    @property
    def charge(self):
        """Formal charge."""
        chem = self.get("chemical_data") or {}
        return chem.get("charge")

    @property
    def chebiAsciiName(self):
        """ASCII name of the compound (primary name)."""
        return self.get("ascii_name") or self.get("name")

    @property
    def chebiId(self):
        """ChEBI accession string, e.g. 'CHEBI:27732'."""
        return self.get("chebi_accession") or str(self.get("id", ""))

    @property
    def DatabaseLinks(self):
        """List of ``(accession_number, source_name)`` tuples from all
        database cross-references, mirroring the old SOAP interface."""
        db_accessions = self.get("database_accessions") or {}
        links = []
        for acc_list in db_accessions.values():
            if isinstance(acc_list, list):
                for acc in acc_list:
                    acc_num = acc.get("accession_number", "")
                    src_name = acc.get("source_name", "")
                    if acc_num or src_name:
                        links.append((acc_num, src_name))
        return links


class ChEBI(REST):
    """Interface to the `ChEBI <https://www.ebi.ac.uk/chebi/>`_ REST API.

    ChEBI (Chemical Entities of Biological Interest) is a freely available
    dictionary of molecular entities focused on 'small' chemical compounds.

    The REST API is documented at
    https://www.ebi.ac.uk/chebi/backend/api/docs/

    Example usage::

        >>> from bioservices import ChEBI
        >>> ch = ChEBI()
        >>> res = ch.getCompleteEntity("CHEBI:27732")
        >>> res.smiles
        'Cn1cnc2c1c(=O)n(c(=O)n2C)C'

    """

    _url = "https://www.ebi.ac.uk/chebi/backend/api/public"

    def __init__(self, verbose=False, cache=False):
        """.. rubric:: Constructor

        :param bool verbose:
        :param bool cache:
        """
        super(ChEBI, self).__init__(name="ChEBI", url=ChEBI._url, verbose=verbose, cache=cache)

    # ------------------------------------------------------------------
    # Internal helpers
    # ------------------------------------------------------------------

    def _chebi_num(self, chebiId):
        """Return the numeric part of a ChEBI identifier.

        Accepts both ``"CHEBI:27732"`` and ``"27732"`` (or int ``27732``).
        """
        s = str(chebiId).strip()
        if ":" in s:
            return s.split(":")[-1]
        return s

    # ------------------------------------------------------------------
    # Public API
    # ------------------------------------------------------------------

    def getCompleteEntity(self, chebiId):
        """Retrieve the complete entity for a ChEBI identifier.

        :param str chebiId: a valid ChEBI identifier (e.g. ``"CHEBI:27732"``)
        :return: a :class:`ChebiEntity` dict-like object

        ::

            >>> from bioservices import ChEBI
            >>> ch = ChEBI()
            >>> res = ch.getCompleteEntity("CHEBI:27732")
            >>> float(res.mass)
            194.19076

        .. seealso:: :meth:`conv`, :meth:`getCompleteEntityByList`
        """
        cid = self._chebi_num(chebiId)
        res = self.http_get("compound/{}".format(cid), frmt="json")
        if isinstance(res, dict):
            return ChebiEntity(res)
        return res

    def conv(self, chebiId, target):
        """Return the cross-reference accession number(s) for a given database.

        Calls :meth:`getCompleteEntity` internally and filters the
        ``DatabaseLinks`` by *target*.

        :param str chebiId: a valid ChEBI identifier (e.g. ``"CHEBI:10102"``)
        :param str target: source database name
            (e.g. ``"KEGG COMPOUND accession"``)
        :return: list of accession number strings

        ::

            >>> ch.conv("CHEBI:10102", "KEGG COMPOUND accession")
            ['C07484']

        """
        res = self.getCompleteEntity(chebiId)
        if not isinstance(res, dict):
            raise ValueError("Could not retrieve entity for %s" % chebiId)

        db_accessions = res.get("database_accessions") or {}
        all_sources = set()
        matches = []
        for acc_list in db_accessions.values():
            if isinstance(acc_list, list):
                for acc in acc_list:
                    src = acc.get("source_name", "")
                    all_sources.add(src)
                    if src == target:
                        matches.append(str(acc.get("accession_number", "")))

        if not matches:
            raise ValueError("valid database target are %s" % sorted(all_sources))
        return matches

    def getLiteEntity(self, search, searchCategory="ALL", maximumResults=200, stars="ALL"):
        """Retrieve a list of lite entities matching a search term.

        :param str search: search string (ChEBI name, identifier, SMILES, etc.)
        :param str searchCategory: filter category (default ``"ALL"``)
        :param int maximumResults: maximum number of results (default 200)
        :param str stars: star filter – ``"ALL"``, ``"TWO ONLY"``,
            or ``"THREE ONLY"`` (default ``"ALL"``)
        :return: list of :class:`ChebiEntity` objects

        ::

            >>> res = ch.getLiteEntity("caffeine", maximumResults=10)
            >>> len(res)
            10

        .. seealso:: :meth:`getCompleteEntity`
        """
        params = {"term": search, "size": maximumResults}
        res = self.http_get("es_search/", frmt="json", params=params)
        if isinstance(res, dict) and "results" in res:
            return [ChebiEntity(r.get("_source") or {}) for r in res["results"]]
        return []

    def getUpdatedPolymer(self, chebiId):
        """Return compound data for a polymer ChEBI entry.

        In the REST API this is equivalent to :meth:`getCompleteEntity`.

        :param str chebiId: a valid ChEBI identifier (string)
        :return: a :class:`ChebiEntity` dict-like object
        """
        return self.getCompleteEntity(chebiId)

    def getCompleteEntityByList(self, chebiIdList=None):
        """Retrieve complete entities for a list of ChEBI identifiers.

        :param list chebiIdList: list of ChEBI identifiers
            (maximum 50 entries recommended)
        :return: list of :class:`ChebiEntity` objects

        .. seealso:: :meth:`getCompleteEntity`
        """
        if chebiIdList is None:
            chebiIdList = []
        results = []
        for cid in chebiIdList:
            entity = self.getCompleteEntity(cid)
            if entity is not None:
                results.append(entity)
        return results

    def getOntologyParents(self, chebiId):
        """Retrieve the ontology parents of a ChEBI entity.

        :param str chebiId: a valid ChEBI identifier (string)
        :return: dict with ontology parent information
        """
        cid = self._chebi_num(chebiId)
        res = self.http_get("ontology/parents/{}/".format(cid), frmt="json")
        return res

    def getOntologyChildren(self, chebiId):
        """Retrieve the ontology children of a ChEBI entity.

        :param str chebiId: a valid ChEBI identifier (string)
        :return: dict with ontology children information
        """
        cid = self._chebi_num(chebiId)
        res = self.http_get("ontology/children/{}/".format(cid), frmt="json")
        return res

    def getAllOntologyChildrenInPath(self, chebiId, relationshipType, onlyWithChemicalStructure=False):
        """Retrieve ontology children connected by a specific relationship type.

        :param str chebiId: a valid ChEBI identifier (string)
        :param str relationshipType: one of ``"is a"``, ``"has part"``,
            ``"has role"``, ``"is conjugate base of"``,
            ``"is conjugate acid of"``, ``"is tautomer of"``,
            ``"is enantiomer of"``, ``"has functional parent"``,
            ``"has parent hydride"``, ``"is substituent group of"``
        :param bool onlyWithChemicalStructure: filter to entities with a
            chemical structure (default ``False``)
        :return: list of ontology relation dicts

        ::

            >>> ch.getAllOntologyChildrenInPath("CHEBI:27732", "has part")

        """
        self.devtools.check_param_in_list(
            relationshipType,
            list(_RELATION_TYPE_MAP.keys()),
        )
        rel_type = _RELATION_TYPE_MAP[relationshipType]
        cid = self._chebi_num(chebiId)
        res = self.http_get("ontology/children/{}/".format(cid), frmt="json")
        if isinstance(res, dict):
            ontology = res.get("ontology_relations") or {}
            incoming = ontology.get("incoming_relations") or []
            filtered = [r for r in incoming if r.get("relation_type") == rel_type]
            return filtered
        return res

    def getStructureSearch(
        self,
        structure,
        mode="MOLFILE",
        structureSearchCategory="SIMILARITY",
        totalResults=50,
        tanimotoCutoff=0.25,
    ):
        """Perform a substructure, similarity, or identity search.

        :param str structure: input structure string
        :param str mode: structure format – ``"MOLFILE"``, ``"SMILES"``,
            or ``"CML"``
        :param str structureSearchCategory: search type –
            ``"SIMILARITY"``, ``"SUBSTRUCTURE"``, or ``"IDENTITY"``
        :param int totalResults: maximum number of results (default 50)
        :param float tanimotoCutoff: minimum Tanimoto score (default 0.25,
            only used for ``"SIMILARITY"`` searches)
        :return: list of matching entities

        ::

            >>> ch = ChEBI()
            >>> smiles = ch.getCompleteEntity("CHEBI:27732").smiles
            >>> ch.getStructureSearch(smiles, "SMILES", "SIMILARITY", 3, 0.25)

        """
        self.devtools.check_param_in_list(structureSearchCategory, ["SIMILARITY", "SUBSTRUCTURE", "IDENTITY"])
        self.devtools.check_param_in_list(mode, ["MOLFILE", "SMILES", "CML"])

        _type_map = {"SMILES": "smiles", "MOLFILE": "mol", "CML": "cml"}
        _cat_map = {"SIMILARITY": "similarity", "SUBSTRUCTURE": "substructure", "IDENTITY": "connectivity"}

        params = {
            "structure": structure,
            "type": _type_map[mode],
            "searchCategory": _cat_map[structureSearchCategory],
            "total": totalResults,
            "tanimoto": tanimotoCutoff,
        }
        res = self.http_get("structure_search/", frmt="json", params=params)
        return res



#
#  This file is part of bioservices software
#
#  Copyright (c) 2013-2014 - EBI-EMBL
#
#  File author(s):
#      Thomas Cokelaer <cokelaer@ebi.ac.uk>
#
#
#  Distributed under the GPLv3 License.
#  See accompanying file LICENSE.txt or copy at
#      http://www.gnu.org/licenses/gpl-3.0.html
#
#  website: https://github.com/cokelaer/bioservices
#  documentation: http://packages.python.org/bioservices
#
##############################################################################
"""Interface to HUGO/HGNC web services

.. topic:: What is HGNC ?

    :URL: http://www.genenames.org
    :Citation:

    .. highlights::

        "The HUGO Gene Nomenclature Committee (HGNC) has assigned unique gene symbols and
        names to over 37,000 human loci, of which around 19,000 are protein coding.
        genenames.org is a curated online repository of HGNC-approved gene nomenclature
        and associated resources including links to genomic, proteomic and phenotypic
        information, as well as dedicated gene family pages."

        -- From HGNC web site, July 2013


"""
from bioservices import REST
from bioservices.xmltools import bs4
import easydev
from bioservices import logger

logger.name = __name__


from urllib.error import HTTPError


__all__ = ["HGNC"]


class HGNC:
    """Wrapper to the genenames web service


    See details at http://www.genenames.org/help/rest-web-service-help

    """

    def __init__(self, verbose=False, cache=False):
        url = "http://rest.genenames.org/"
        self.services = REST("HGNC", url=url, verbose=verbose, cache=cache)

        self._info = self.get_info()
        self.searchable_fields = self._info["searchableFields"]
        self.stored_fields = self._info["storedFields"]

    def get_info(self, frmt="json"):
        """Request information about the service

        Fields are when the server was last updated (lastModified),
        the number of documents (numDoc), which fields can be queried
        using search and fetch (searchableFields) and which fields may
        be returned by fetch (storedFields).


        """
        headers = self.services.get_headers(content=frmt)
        res = self.services.http_get("info", frmt=frmt, headers=headers)
        return res

    def fetch(self, database, query, frmt="json"):
        """Retrieve particular records from a searchable fields

        Returned object is a json object with fields as in
        :attr:`stored_field`, which is returned from :meth:`get_info` method.

        Only one query at a time. No wild cards are accepted.
        ::

            >>> h = HGNC()
            >>> h.fetch('symbol', 'ZNF3')
            >>> h.fetch('alias_name', 'A-kinase anchor protein, 350kDa')
        """
        easydev.check_param_in_list(database, self.searchable_fields)
        url = "fetch/{0}/{1}".format(database, query)
        headers = self.services.get_headers(content=frmt)
        res = self.services.http_get(url, frmt=frmt, headers=headers)
        return res

    def search(self, database_or_query=None, query=None, frmt="json"):
        """Search a searchable field (database) for a pattern

        The search request is more powerful than fetch for querying the
        database, but search will only returns the fields hgnc_id, symbol and
        score. This is because this tool is mainly intended to query the server
        to find possible entries of interest or to check data (such as your own
        symbols) rather than to fetch information about the genes. If you want
        to retrieve all the data for a set of genes from the search result, the
        user could use the hgnc_id returned by search to then fire off a fetch
        request by hgnc_id.

        :param database: if not provided, search all databases.


        ::

            # Search all searchable fields for the tern BRAF
            h.search('BRAF')

            # Return all records that have symbols that start with ZNF
            h.search('symbol', 'ZNF*')

            # Return all records that have symbols that start with ZNF
            # followed by one and only one character (e.g. ZNF3)
            # Nov 2015 does not work neither here nor in within in the
            # official documentation
            h.search('symbol', 'ZNF?')

            # search for symbols starting with ZNF that have been approved
            # by HGNC
            h.search('symbol', 'ZNF*+AND+status:Approved')

            # return ZNF3 and ZNF12
            h.search('symbol', 'ZNF3+OR+ZNF12')

            # Return all records that have symbols that start with ZNF which
            # are not approved (ie entry withdrawn)
            h.search('symbol', 'ZNF*+NOT+status:Approved')

        """
        if database_or_query is None and query is None:
            raise ValueError("you must provide at least one parameter")
        elif database_or_query is not None and query is None:
            # presumably user wants to search all databases
            query = database_or_query
            url = "search/{0}".format(query)
        else:
            database = database_or_query
            easydev.check_param_in_list(database, self.searchable_fields)
            url = "search/{0}/{1}".format(database, query)

        headers = self.services.get_headers(content=frmt)
        res = self.services.http_get(url, frmt=frmt, headers=headers)
        return res



#
#  This file is part of BioServices
#
#  Distributed under the terms of the 3-clause BSD license.
#  The full license is in the LICENSE file, distributed with this software.
#
#  website: https://github.com/sequana/sequana
#  documentation: http://sequana.readthedocs.io
#
##############################################################################
import functools
import glob
import os
import subprocess
from pathlib import Path

import colorlog
import rich_click as click

logger = colorlog.getLogger(__name__)

__all__ = ["main"]


# This can be used by all commands as a simple decorator
def common_logger(func):
    @click.option(
        "--logger",
        default="INFO",
        type=click.Choice(["INFO", "DEBUG", "WARNING", "CRITICAL", "ERROR"]),
    )
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        return func(*args, **kwargs)

    return wrapper


CONTEXT_SETTINGS = dict(help_option_names=["-h", "--help"])


from bioservices import version


@click.group(context_settings=CONTEXT_SETTINGS)
@click.version_option(version=version)
def main(**kwargs):
    """This is the main entry point for a set of BioServices applications"""
    pass


@main.command()
@click.option("--accession", type=click.STRING, required=True, help="A valid accession number (e.g., FN433596.1)")
@click.option(
    "--prefix",
    type=click.STRING,
    default=None,
    help="""By default, output FastA file is name after the accession number
            but you can rename it using this --prefix name""",
)
@click.option("--method", type=click.Choice(["ENA", "EUtils"]), default="EUtils")
@click.option("--with-gff3", is_flag=True, default=False)
@click.option("--with-gbk", is_flag=True, default=False)
def download_accession(**kwargs):
    """Download Fasta related to an accession and possibly other type (e.g gff)

    Input file can be gzipped or not. The --output-file

        bioservices download-accession FN433596.1
    """
    from bioservices.apps.download_fasta import download_fasta
    from bioservices.apps.download_gbk import download_gbk
    from bioservices.apps.download_gff3 import download_gff3

    prefix = kwargs["prefix"]

    logger.info("Downloading FastA file")
    download_fasta(kwargs["accession"], output_filename=f"{prefix}.fa" if prefix else prefix, method=kwargs["method"])

    if kwargs["with_gff3"]:
        logger.info("Downloading GFF file")
        download_gff3(
            kwargs["accession"], output_filename=f"{prefix}.gff" if prefix else prefix, method=kwargs["method"]
        )

    if kwargs["with_gbk"]:
        logger.info("Downloading Genbank file")
        download_gbk(
            kwargs["accession"], output_filename=f"{prefix}.gbk" if prefix else prefix, method=kwargs["method"]
        )


@main.command()
# @click.option("--name", type=click.STRING, help="A valid accession number (e.g., FN433596.1)")
@click.option("--id", type=click.STRING, help="A valid taxon ID (e.g., 9606)")
@click.option(
    "--method", type=click.STRING, default="EUtils", help="A method. EUtils only option implemented right now"
)
def taxonomy(**kwargs):
    """Download information related to a taxid

    bioservices taxonomy --id 9606
    """

    if kwargs["method"] == "EUtils":
        from bioservices import EUtils

        eu = EUtils(email="bioservices")
        ret = eu.ESummary("taxonomy", kwargs["id"])
        for uid in ret["uids"]:
            print(ret[uid])
